#include "src/common.h"
#include "src/pkcs7.h"
#include "src/enc_modes.h"
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <openssl/evp.h>
#include <openssl/rand.h>
#include <assert.h>

struct SyncOperations
{
    int (*init)( EVP_CIPHER_CTX *ctx, const EVP_CIPHER *cipher,
                    const unsigned char *key, const unsigned char *iv);
    int (*update)(EVP_CIPHER_CTX *ctx, unsigned char *out, int *outl,
                    const unsigned char *in, int inl);
    int (*final)(EVP_CIPHER_CTX *ctx, unsigned char *out, int *outl);
};

struct SyncOperations EncryptOps = {
    .init = EVP_EncryptInit,
    .update = EVP_EncryptUpdate,
    .final = EVP_EncryptFinal
};

struct SyncOperations DecryptOps = {
    .init = EVP_DecryptInit,
    .update = EVP_DecryptUpdate,
    .final = EVP_DecryptFinal
};


struct OpenSSL
{
    static uint8_t  Cbc(CryptoAttribs_t* i_attribs,
                        const Key_t* i_key)
    {
        struct SyncOperations sop = EncryptOps;
        if(i_attribs->operation == kDecrypt)
            sop = DecryptOps;

        if(NULL==i_attribs->output)
        {
            i_attribs->output = (uint8_t*) malloc(i_attribs->input_len);
        }

        int ret=0;
        EVP_CIPHER_CTX ctx;
        EVP_CIPHER_CTX_init(&ctx);
        sop.init(&ctx, EVP_aes_128_cbc(), i_key->key, i_attribs->iv);
        EVP_CIPHER_CTX_set_padding(&ctx, 0);
        EVP_CIPHER_CTX_set_key_length(&ctx, i_key->len);

        sop.update(&ctx, i_attribs->output, &ret, i_attribs->input,
            i_attribs->input_len);
        i_attribs->output_len = ret;
        sop.final (&ctx, (i_attribs->output)+ret, &ret);
        i_attribs->output_len += ret;
        return 1;
    }
};

TCASE(ecb_encrypt_decrypt_single_block)
{
    static const uint8_t expected_result[17] = "0123456789123456";
    CryptoAttribs_t encode, decode;
    Key_t key1;
    CryptoAttribs_t::Init(&encode);
    CryptoAttribs_t::Init(&decode);
    Key_t::Init(&key1);

    encode.input = (uint8_t*)malloc(16);
    encode.input_len=16;
    decode.input = (uint8_t*)malloc(16);
    decode.input_len=16;
    key1.key = (uint8_t*)malloc(16);
    key1.len = 16;

    memcpy(key1.key, "YELLOW SUBMARINE", 16);
    memcpy(encode.input, expected_result, 16);
    ecb_encrypt(&encode, &key1);
    CHECK(encode.output_len==16);

    memcpy(decode.input, encode.output, 16);
    ecb_decrypt(&decode, &key1);
    CHECK(decode.output_len == 16);

    CHECK(memcmp(decode.output, expected_result, 16) == 0);

    CryptoAttribs_t::Free(&encode);
    CryptoAttribs_t::Free(&decode);
    Key_t::Free(&key1);
}
TCASE_E

TCASE(pkcs7_test)
{
    // int pkcs7_pad(const char* i_buff, const size_t i_len, char* o_buff );
    const char text1[] = "Text1";
    const char text2[] = "Text2ToPad";
    int ret;
    char buff1[32];
    char buff2[10];

    ret = pkcs7_pad(text1, 5, buff1, 32);
    CHECK(ret==0, (const unsigned char* const) "Padding operation failed");
    ret = memcmp(text1, buff1, 5);
    CHECK(ret==0, (const unsigned char* const) "Content differs");
    for(int i=5; i<32; i++)
    {
        CHECK(buff1[i] == 27, (const unsigned char* const) "Wrong padding");
    }
}
TCASE_E

TCASE(cbc_enc_dec_test)
{
    static const uint8_t test_text[49] = "The quick brown fox jumps over the lazy mad dog.";
    CryptoAttribs_t enc_attribs;
    CryptoAttribs_t dec_attribs;
    Key_t key;
    CryptoAttribs_t::Init(&enc_attribs);
    CryptoAttribs_t::Init(&dec_attribs);
    Key_t::Init(&key);

    // Setup key
    key.len = 16;
    key.key = (uint8_t*) malloc(key.len);
    memcpy(key.key, "0123456789123456", key.len);

    // Encrypt
    enc_attribs.input_len = sizeof(test_text)-1;
    enc_attribs.input = (uint8_t*) malloc(enc_attribs.input_len);
    enc_attribs.operation = kEncrypt;
    memcpy( enc_attribs.input, test_text, enc_attribs.input_len);
    enc_attribs.iv_len = key.len;
    enc_attribs.iv = (uint8_t*)malloc(enc_attribs.iv_len);
    memset(enc_attribs.iv, 0, enc_attribs.iv_len);
    cbc_encrypt(&enc_attribs, &key);

    // Setup decryption object.
    dec_attribs.iv = (uint8_t*)malloc( enc_attribs.iv_len );
    dec_attribs.iv_len = enc_attribs.iv_len;
    dec_attribs.operation = kDecrypt;
    memset(dec_attribs.iv, 0, dec_attribs.iv_len);
    dec_attribs.input = (uint8_t*)malloc(enc_attribs.output_len);
    dec_attribs.input_len = enc_attribs.output_len;
    memcpy(dec_attribs.input, enc_attribs.output, enc_attribs.output_len);

    // Decrypt
    cbc_decrypt(&dec_attribs, &key);
    CHECK( dec_attribs.output_len+1 == sizeof(test_text), (uint8_t*) "wrong size of ouptut");
    CHECK(memcmp(dec_attribs.output, test_text, dec_attribs.output_len) == 0, dec_attribs.output);

end:
    CryptoAttribs_t::Free(&enc_attribs);
    CryptoAttribs_t::Free(&dec_attribs);
    Key_t::Free(&key);
}
TCASE_E

TCASE(cbc_decrypt_test)
{
    static const uint8_t expected_result[34] = "I'm back and I'm ringin' the bell";
    CryptoAttribs_t attribs;
    Key_t key;
    Result_t res = Result_Error;

    Key_t::Init(&key);
    key.len = 16;
    key.key = (uint8_t*) malloc(key.len);
    memcpy(key.key,"YELLOW SUBMARINE",key.len);

    CryptoAttribs_t::Init(&attribs);
    attribs.iv = (uint8_t*) malloc(key.len);
    attribs.iv_len = key.len;
    memset(attribs.iv, 0, attribs.iv_len);

    res = load_base64_to_hex(
                "sol/etc/set2_t2.txt",
                &(attribs.input),
                &(attribs.input_len));
    CHECK(res == Result_OK, (const uint8_t*) "Problem when loading input file");
    cbc_decrypt(&attribs, &key);
    CHECK( memcmp(expected_result, attribs.output, 33) == 0, (const uint8_t*)
        "Wrong plaintext decrypted");
    // cleanup
    CryptoAttribs_t::Free(&attribs);
    Key_t::Free(&key);
}
TCASE_E

/*
TCASE(encode_decode_openssl)
{
    uint8_t     concatenated_blocks[16*3];
    uint8_t     iv1[16]  = {0};
    uint8_t     key[17] = "YELLOW SUBMARINE"; // 16+'\0' = 17
    uint8_t*    ciphertext = NULL;
    size_t      ciphertext_len = 0;
    uint8_t*    out = NULL;
    size_t      out_len = 0;
    Result_t    res = Result_Error;
    memset(concatenated_blocks, 2, 16);
    memset(&concatenated_blocks[16], 5, 16);
    memset(&concatenated_blocks[32], 9, 16);

    RAND_bytes(iv1, 16);

    CHECK( 1 == OpenSSL::Cbc(concatenated_blocks, 48,
                            iv1, 16,
                            key, 16,
                            kEncrypt,
                            &ciphertext, &ciphertext_len),
            (uint8_t*)"OpenSSL encryption failed");

    // test decryption
    CHECK(ciphertext_len==48, (uint8_t*)"Ciphertext has wrong size");
    res = decrypt_cbc(iv1, 16, ciphertext, ciphertext_len, key, 16, &out, &out_len );
    CHECK( Result_OK == res );
    CHECK( memcmp(concatenated_blocks, out, 48) == 0, (uint8_t*)"Input/Output differs");

    // test encryption
    ciphertext_len = 0;
    out_len = 0;
    res = encrypt_cbc(iv1, 16, concatenated_blocks, 48, key, 16, &ciphertext, &ciphertext_len );
    CHECK( Result_OK == res );
    CHECK(ciphertext_len==48, (uint8_t*)"Ciphertext has wrong size");
    CHECK( 1 == OpenSSL::Cbc(ciphertext, 48,
                            iv1, 16,
                            key, 16,
                            kDecrypt,
                            &out, &out_len),
            (uint8_t*)"OpenSSL encryption failed");
    CHECK(ciphertext_len==48, (uint8_t*)"Ciphertext has wrong size");
    CHECK( memcmp(concatenated_blocks, out, 48) == 0, (uint8_t*)"Input/Output differs");
}
TCASE_E
*/