#include <stdint.h>
#include <string.h>
#include <inttypes.h>
#include <openssl/evp.h>
#include <assert.h>
#include "common.h"
#include "xor.h"
#include "base64.h" // to del
#include <openssl/err.h>

#define OP_CHECK(exp)                                           \
    do {                                                        \
        if(!exp) {                                              \
            printf("Error with expression %s %d\n", #exp, exp); \
            goto end;                                           \
        }                                                       \
    }                                                           \
    while(false)

Result_t ecb_decrypt(
    CryptoAttribs_t* attribs,
    const Key_t* const key)
{
    // if output is NULL then allocate same size as input rounded to block size
    if(attribs->output == NULL)
    {
        attribs->output = (uint8_t*) malloc(
                    ROUNDUP_16_BYTES(attribs->input_len) );
        attribs->output_len = attribs->input_len;
    }
    uint8_t* out = attribs->output;

    int ret = 0;
    EVP_CIPHER_CTX* ctx = EVP_CIPHER_CTX_new();
    EVP_CIPHER_CTX_init(ctx);

    OP_CHECK( EVP_DecryptInit_ex(ctx, EVP_aes_128_ecb(), NULL, key->key, NULL) );
    EVP_DecryptInit_ex(ctx, NULL, NULL, NULL, NULL);
    OP_CHECK(  EVP_CIPHER_CTX_iv_length(ctx) == 0);
    EVP_CIPHER_CTX_set_padding(ctx, 0);
    //OP_CHECK( EVP_CIPHER_CTX_set_key_length(ctx, key->len) );
    OP_CHECK( EVP_DecryptUpdate(ctx, &out[ret], &ret, attribs->input, attribs->input_len) );
    attribs->output_len = ret;
    OP_CHECK( EVP_DecryptFinal_ex(ctx, &out[ret], &ret) );
    attribs->output_len += ret;
    EVP_CIPHER_CTX_free(ctx);
//    EVP_cleanup();

    return Result_OK;

end:
    return Result_Error;
}

Result_t ecb_encrypt(
    CryptoAttribs_t* attribs,
    const Key_t* const key)

{
    assert(attribs->input_len == key->len);
    assert(key->len == 16);
    // if output is NULL then allocate same size as input rounded to block size
    if(attribs->output == NULL)
    {
        attribs->output =  (uint8_t*) malloc(
                    ROUNDUP_16_BYTES(attribs->input_len) );
    }

    uint8_t* out = attribs->output;
    int ret = 0;

    EVP_CIPHER_CTX* ctx = EVP_CIPHER_CTX_new();
    EVP_CIPHER_CTX_init(ctx);
    OP_CHECK( EVP_EncryptInit_ex(ctx, EVP_aes_128_ecb(), NULL, key->key, NULL) );
        EVP_EncryptInit_ex(ctx, NULL, NULL, NULL, NULL);

    OP_CHECK(  EVP_CIPHER_CTX_iv_length(ctx) == 0);
    EVP_CIPHER_CTX_set_padding(ctx, 0);
    //OP_CHECK( EVP_CIPHER_CTX_set_key_length(ctx, key->len) );
    OP_CHECK( EVP_EncryptUpdate(ctx, &out[ret], &ret, attribs->input, attribs->input_len) );
    attribs->output_len = ret;
    OP_CHECK( EVP_EncryptFinal_ex(ctx, &out[ret], &ret) );
    attribs->output_len += ret;
    EVP_CIPHER_CTX_free(ctx);
//    EVP_cleanup();
    return Result_OK;
end:
    return Result_Error;
}


Result_t cbc_decrypt(
    CryptoAttribs_t* attribs,
    const Key_t* const key)
{
    if((attribs->input_len % key->len) != 0 )
        return Result_Error;

    if( attribs->iv_len != key->len )
        return Result_Error;

    if( key->len == 0 )
        return Result_Error;

    const size_t bc = attribs->input_len / key->len;

    // Allocate memory for output, if NULL
    if( attribs->output == NULL)
    {
        // same as input + 16 bytes iv
        const size_t len = (ROUNDUP_16_BYTES(attribs->input_len)+16);
        attribs->output = (uint8_t*) malloc(len);
        //attribs->output_len = len;
        assert( attribs->output != NULL );
        assert( len != 0 );
    }

    uint8_t* iv = attribs->iv;
    // Allocate struct that will be used for block decryption
    CryptoAttribs_t ecb_attribs;
    CryptoAttribs_t::Init(&ecb_attribs);
    ecb_attribs.output = (uint8_t*) malloc(key->len);
    ecb_attribs.output_len = key->len;

    for(int i=0; i<bc; ++i)
    {
        ecb_attribs.input = (attribs->input)+(key->len*i);
        ecb_attribs.input_len = key->len;

        // 1. Decrypt
        if( ecb_decrypt(&ecb_attribs, key) != Result_OK )
            return Result_Error;

        // 2. Xor IV
        unsigned char* out = attribs->output+(key->len)*i;
        xor_repeatedly(
            iv, attribs->iv_len,
            ecb_attribs.output, ecb_attribs.output_len,
            out);

        // 3. Change iv with block
        iv = ecb_attribs.input;
        attribs->output_len += key->len;
    }
    free(ecb_attribs.output);
    return Result_OK;
}

Result_t cbc_encrypt(
    CryptoAttribs_t* attribs,
    const Key_t* const key)
{
    if((attribs->input_len % key->len) != 0 )
        return Result_Error;

    if( attribs->iv_len != key->len )
        return Result_Error;

    if( key->len == 0 )
        return Result_Error;

    const size_t bc = attribs->input_len / key->len;
    const size_t bs = key->len;

    if( attribs->output == NULL)
    {
        // same as input + 16 bytes iv
        attribs->output =  (uint8_t*) malloc(
                    (ROUNDUP_16_BYTES(attribs->input_len)+16) );
        assert( attribs->output != NULL );
    }

    uint8_t* iv = attribs->iv;
    // Initialize attribs for ECB block encryption
    CryptoAttribs_t ecb_attribs;
    CryptoAttribs_t::Init(&ecb_attribs);
    ecb_attribs.input = (uint8_t*)malloc(bs);
    ecb_attribs.input_len = bs;
    ecb_attribs.output = (uint8_t*)malloc(bs);
    ecb_attribs.output_len = bs;
    for(int i=0; i<bc; ++i)
    {
        uint8_t* buffptr = (attribs->input)+(bs*i);

        // 1. Xor IV
        xor_repeatedly(
            iv, bs,
            buffptr, bs,
            ecb_attribs.input);
        ecb_attribs.input_len = bs;

        // 2. Encrypt
        if( ecb_encrypt(&ecb_attribs, key) != Result_OK )
            return Result_Error;

        // Encrypted block is my new IV
        iv = ecb_attribs.output;

        // 3. Change iv with block
        memcpy((attribs->output)+(bs*i),
                ecb_attribs.output,
                ecb_attribs.output_len);
        attribs->output_len += ecb_attribs.output_len;
    }
    CryptoAttribs_t::Free(&ecb_attribs);
    return Result_OK;
}
